/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};

document.addEventListener('gc-databind-ready', () => {

    var numDevices = 4;
    var partialMessage = [];
    var uartPacketLength = 0;
    var registerInfo;

    var tempRegCache;
    var configRegCache;
    var lowRegCache;
    var highRegCache;

    var readWriteDropdown;
    var registerPage;
    var registerModel;

    var isRead = true;

    var uartDeferred = undefined;
    let seq = Promise.resolve();
    let queue = [];
    let guhbusy = false;
    var UARTPacketCodec = function () {
        this.debounceWriteValue = UARTPacketCodec.prototype.debounce(this.doWriteValue.bind(this)); 
    };
    UARTPacketCodec.prototype = new gc.databind.IPollingPacketCodec();
    UARTPacketCodec.prototype.sendConvertTemp = true;
    UARTPacketCodec.prototype.map = {};
    UARTPacketCodec.prototype.cmd = {
        ReadROM: 0x33,
        MatchROM: 0x55,
        SearchROM: 0xF0,
        SkipROM: 0xCC,
        ConvertT: 0x44,
        CopyScratchPad2: 0x55,
        ReadMemory: 0xF0,
        WriteScratchPad1: 0x4E,
        WriteScratchPad2: 0x0F,
        ReadScratchPad2: 0xAA,
        ReadScratchPad1: 0xBE,
        ReadPowerSupply: 0xB4,
        ReadEEPROM: 0xF0
    }
    UARTPacketCodec.prototype.json = {
        ScratchPad1: 'ScratchPad-1',
        ScratchPad2: 'ScratchPad-2'
    }
    UARTPacketCodec.prototype.reflect8 = function (msg) {
        let out = 0;
        for (bit = 0; bit < 8; bit++) {
            out |= msg & (1 << bit) ? (1 << (7 - bit)) : 0;
        }
        return out;
    }
    UARTPacketCodec.prototype.crc = function (msg) {
        let crc = 0x00;
        for (i = 0; i < msg.length; i++) {
            crc ^= this.reflect8(msg[i]);
            for (bit = 0; bit < 8; bit++) {
                if (crc & 0x80)
                    crc = (crc << 1) ^ 0x31;
                else
                    crc = (crc << 1);
            }
        }
        return this.reflect8(crc) & 0xFF;
    };

    /**
     * Decodes packets from the target device into data objects.  One object for each packet of data.
     * Unless you have chained your custom codec with the CR codec, the packet data is not framed,
     * so this method is responsible for decoding partial packets as well as multiple packets.  Partial packets
     * should not be returned, but deferred until the next decode method call with more raw data to process.  As a
     * result, this method may return 0 or more packets of data on each invocation.
     *
     * @param target -  function to call to pass on the encoded data towards the GUI
     * @param data  - The raw data received that needs to be decoded from packets into objects.
     * @return - true if connected (i.e. valid data received), false if not connected
     */
    UARTPacketCodec.prototype.decode = function (target, data) {
        if (data & 1)
            // check if the first bit transmitted is low to filter bad packets
            console.log("Warning: decode ignoring bad packet: 0x" + Number(data).toString(16))
        else
            partialMessage = partialMessage.concat(data);
        // console.log("uartPacketLength", uartPacketLength);
        // console.log("partialMessage:", partialMessage);
        if (partialMessage.length >= uartPacketLength) {
            let that = this;
            let dmsg = "";
            let bit = 0;
            let bytes = Array();
            let val = 0;
            partialMessage.forEach((x, i) => {
                if (bit > 7) {
                    bit = 0;
                    bytes.push(val);
                    val = 0;
                }
                if (x & 1) val += 2 ** bit;
                bit++;
            });
            bytes.push(val);
            bytes.forEach((x, i) => dmsg += i.toString() + ": 0x" + Number(x).toString(16) + "\n");
            console.log(dmsg);
            if (this.partialPacketQ.length > 0) {
                if (this.partialPacketQ[0].len == bytes.length)
                    this.partialPacketQ.shift().p.resolve(bytes);
            } else {
                let cmd1 = bytes.splice(0, 1);
                if (cmd1 == this.cmd.SkipROM) {
                    let cmd2 = bytes.splice(0, 1);
                    let crc = bytes.pop();
                    if (cmd2 == this.cmd.ReadScratchPad1) {
                        let crc1 = bytes.splice(8, 1);
                        bytes.forEach((x, i) => this.map[this.json.ScratchPad1][i] ? this.map[this.json.ScratchPad1][i].resolve([x]) : undefined);
                        this.map[this.json.ScratchPad1] = undefined;
                    } else if ((cmd2 == this.cmd.ReadScratchPad2) || (cmd2 == this.cmd.ReadEEPROM)) {
                        this.scratchpad2Q.shift().resolve(bytes);
                    }
                }
            }
            guhbusy = false;
            if (queue.length > 0) {
                const next = queue.shift();
                this.guh(next.baud, next.data, next.bypass);
            }
            partialMessage = [];
        }
    };

    UARTPacketCodec.prototype.codecReset = function () {
        guhbusy = false;
        queue = [];
        partialMessage = [];
        this.map = {};
    }
    UARTPacketCodec.prototype.ren = function () {
        console.log("ren");
        this.OWReset();
        this.OWTransfer([0xCC, 0xBE], 9); //skiprom, scratch1
        this.OWReset();
        this.OWTransfer([0xCC, 0xBE], 9); //skiprom, scratch1
    }
    UARTPacketCodec.prototype.guh = function (baud, data, bypass) {
        const u2a = document.querySelector('ti-model-register').getCodec();
        u2a.sendCommandPacket(U2A.UART_Control, [baud, 0, 0, 0, 0]);
        uartPacketLength = data.length;
        while (data.length > 0) this.encoder(data.splice(0, 54));
        if (!bypass)
            guhbusy = true;
        else {
            if (queue.length > 0) {
                const next = queue.shift();
                this.guh(next.baud, next.data, next.bypass);
            }
        }
    }
    UARTPacketCodec.prototype.OWEncode = function (baud, data, bypass = false) {
        if ((queue.length > 0) | guhbusy)
            queue.push({ baud: baud, data: data, bypass: bypass });
        else
            this.guh(baud, data, bypass);
    }
    UARTPacketCodec.prototype.OWReset = function () {
        let bypassBusy = false;
        this.OWEncode(U2A.Baud.b9600, [0xF0], bypassBusy);
    }
    UARTPacketCodec.prototype.OWTransfer = function (data, readBytes, reset = true) {
        let that = this;
        let a = Array();
        data.forEach(d => {
            for (let i = 0; i < 8; i++) {
                a.push((d & (2 ** i)) ? 0xFF : 0xC0);
            }
        })
        for (let i = 0; i < (readBytes * 8); i++) {
            a.push(0xFF);
        }
        if (reset)
            this.OWReset();
        this.OWEncode(U2A.Baud.b115200, a)
    }
    UARTPacketCodec.prototype.OWRegRead = function (grp) {
        if (grp == this.json.ScratchPad1) //Register.json value
            this.OWTransfer([0xCC, 0xBE], 18);
        else if (grp == this.json.ScratchPad2)
            this.OWTransfer([0xCC, 0xAA, 0x00], 9);
    }
    /**
     * Initializes target communication and assigns addresses to each device
     *
     * @param settings - UART interface specific settings
     * @return - promise that resolves once response from target is received
     */
    UARTPacketCodec.prototype.connect = function (settings) {
        this.codecReset();
        registerPage = document.querySelector('ti-widget-register-page');
        readWriteDropdown = document.querySelector('#read_write_dropdown');
        registerModel = document.querySelector('ti-model-register').getModel();

        registerPage.addEventListener('auto_read_config_changed', function (event) {
            var autoRead = +event.detail.autoRead.value;
            if (autoRead !== -1) {
                // Disable read register buttons if polling
                registerPage.$.read_register.disabled = true;
                registerPage.$.read_all_registers.disabled = true;
            }
            else {
                registerPage.$.read_register.disabled = false;
                registerPage.$.read_all_registers.disabled = false;
            }
        });

        return Q(0);
    };

    /**
     * Starts a read operation
     *
     * @param info - identifies which register and which device to read from
     * @return - promise that resolves once response from target is received
     */
    UARTPacketCodec.prototype.readValue = function (info) {
        let grp = info["parentGroup"].name;
        let name = info["fullyQualifiedName"];
        let addr = Number(info["addr"]);
        let that = this;
        console.log("readValue", grp, addr, name);
        if (that.map[grp] === undefined) {
            if (that.sendConvertTemp)
                that.OWTransfer([0xCC, 0x44], 0);
            // need at least 5.7ms delay here, but this implementation gives it for free
            that.OWRegRead(grp);
            that.map[grp] = {};
        }
        that.map[grp][addr] = new Q.defer();
        return that.map[grp][addr].promise.timeout(5e3, "UARTPacketCodec Register Read Timed Out").fail(e => {
            that.codecReset();
            console.log(e)
            return e;
        });
    };
    UARTPacketCodec.prototype.writeValueQ = [];

    UARTPacketCodec.prototype.debounce = function (func, timeout = 10) {
        let timer;
        return (...args) => {
            clearTimeout(timer);
            timer = setTimeout(() => { func.apply(this, args); }, timeout);
        };
    }

    UARTPacketCodec.prototype.doWriteValue = function (info) {
        let grp = info["parentGroup"].name;
        if (grp == this.json.ScratchPad1) {
            let regs = info["parentGroup"]["registers"].map(x => {
                return gc.databind.registry.getBinding(`reg.${x.id}`).getValue();
            });
            let packet = [0xCC, this.cmd.WriteScratchPad1].concat(regs.slice(4, 7), regs.slice(8, 14));
            // packet.push(this.crc(packet));
            this.OWTransfer(packet, 1);
            let x = [...this.writeValueQ];
            this.writeValueQ = [];
            x.forEach(p => p.resolve());
        }
    };
    //this was moved and no longer needed here
    // UARTPacketCodec.prototype.debounceWriteValue = UARTPacketCodec.prototype.debounce(UARTPacketCodec.prototype.doWriteValue.bind(UARTPacketCodec));
    // this.debounceWriteValue = UARTPacketCodec.prototype.debounce(this.doWriteValue.bind(this)); 
    /**
     * Starts a write operation
     *
     * @param info - identifies which register and which device to write to
     * @param value - the value to write to the target
     * @return - promise that resolves once response from target is received
     */
     UARTPacketCodec.prototype.writeValue = function (info, value) {
        let grp = info["parentGroup"].name;
        let name = info["fullyQualifiedName"];
        let addr = Number(info["addr"]);
        console.log("writeValue", grp, addr, name, value);
        this.debounceWriteValue(info);
        let p = new Q.defer();
        this.writeValueQ.push(p);
        return p.promise.timeout(5e3, "UARTPacketCodec Register Write Timed Out").fail(e => {
            that.codecReset();
            console.log(e)
            return e;
        });
    };

    UARTPacketCodec.prototype.scratchpad2Q = [];
    UARTPacketCodec.prototype.partialPacketQ = [];
    UARTPacketCodec.prototype.scratchpad2 = function (data, readBytes, reset = true) {
        console.log("scratchpad2", data);
        this.OWTransfer(data, readBytes, reset);
        let p = new Q.defer();
        if (reset)
            this.scratchpad2Q.push(p);
        else
            this.partialPacketQ.push({ p: p, len: data.length + readBytes })
        return p.promise.timeout(5e3, "UARTPacketCodec ScratchPad2 Timed Out").fail(e => {
            that.codecReset();
            console.log(e)
            return e;
        });
    };

    gc.databind.registerCustomCodec("UARTPacketCodec", UARTPacketCodec, "");
});
