/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};
var saveAsFilePath; // undefined initially
var regFileFormat = "CSV";

const U2A =
{
    I2C_Control: 1, // 0x01
    I2C_Write: 2, // 0x02
    I2C_Read: 3, // 0x03
    I2CRead_WithAddress: 4, // 0x04
    UART_Control: 15, // 0x0F
    UART_Write: 18, // 0x12
    UART_SetMode: 19, // 0x13
    UART_Read: 20, // 0x14
    Baud: {
        b9600: 0,
        b19200: 1,
        b38400: 2,
        b57600: 3,
        b115200: 4,
        b230400: 5,
        b300: 6,
        b320: 7,
        b600: 8,
        b1200: 9,
        b2400: 10,
        b4800: 11
    }
}

const myPlotly = {};
myPlotly.update = (y1) => {
    if (templateObj) {
        const myPlot = templateObj.$.ti_widget_plot;
        if (typeof myPlot.data[1] == 'undefined') {
            const ti_red = '#CC0000';
            const ti_blue = '#115566';
            const not_red = '#000000';
            const not_blue = '#AAAAAA';
            myPlot.data.push({ x: Array(), y: Array() });
            myPlot.data[0].name = '°C';
            myPlot.data[0].line = { color: ti_red };
            myPlot.layout.yaxis = {
                title: '°C',
                range: [0, 100],
                titlefont: { color: ti_red },
                tickfont: { color: ti_red }
            };
        };
        if (myPlot.data[0].x.length > 99) {
            myPlot.data[0].y.shift();

        } else {
            myPlot.data[0].x.push(myPlot.data[0].x.length + 1);
        }
        myPlot.data[0].y.push(y1);
        myPlot.redraw();
    }
}

const FileSaver = {};
const defaultFile = 'mydata.csv';
const fileHeader = 'RH %, Temperature C, Time\n'
const isOnline = typeof nw == 'undefined' || typeof process == 'undefined';
let dbgOnline = false;
if (isOnline | dbgOnline) {
    let mywriter;
    const myTextEnc = new TextEncoder();
    FileSaver.saveStart = function (filename) {
        if (!filename) filename = defaultFile;
        const mystream = streamSaver.createWriteStream(filename);
        FileSaver.saveAbort();
        mywriter = mystream.getWriter();
        FileSaver.saveData(fileHeader);
    };
    FileSaver.saveStop = function () {
        if (mywriter) {
            mywriter.close();
            mywriter = null;
        }
    };
    FileSaver.saveAbort = function () {
        if (mywriter) {
            mywriter.abort('reason');
            mywriter = null;
        }
    };
    FileSaver.saveData = function (data) {
        if (mywriter) {
            mywriter.write(myTextEnc.encode(data));
        }
    };
    window.addEventListener('unload', (event) => {
        FileSaver.saveAbort();
    });
} else {
    // node fs API: https://nodejs.org/api/fs.html
    const fs = require('fs');
    FileSaver.saveLog = false;
    FileSaver.saveFile;
    let saveStarted = undefined;

    FileSaver.saveStart = function (filename) {
        saveStarted = true;
        templateObj.$.file_input.setAttribute('nwsaveas', defaultFile);
        templateObj.$.file_input.click();
    };
    FileSaver.saveStop = function () {
        FileSaver.saveLog = false;
    };
    FileSaver.saveAbort = function () {
        FileSaver.saveLog = false;
    };
    FileSaver.saveData = function (data) {
        if (!FileSaver.saveLog) return;
        if (saveStarted) {
            saveStarted = false;
            fs.writeFileSync(FileSaver.saveFile, fileHeader, { flag: 'w' /* overwrite */ });
        };
        fs.writeFileSync(FileSaver.saveFile, data, { flag: 'a' /* append to file */ });
    };
}
/*
 *  Boilerplate code for creating computed data bindings
 */
document.addEventListener('gc-databind-ready', function () {
    // templateObj.$.ti_widget_button_readROM.addEventListener("click", function (event) {
    //     const u2a = document.querySelector('ti-model-register').getCodec();
    //     u2a.uartInterface.uartCodec.ren()
    // });
});

/*
 *  Boilerplate code for creating custom actions
 */
document.addEventListener('gc-nav-ready', function () {
    /*
     *   Add custom actions for menu items using the following api:
     *
     *   function gc.nav.registryAction(id, runable, [isAvailable], [isVisible]);
     *
     *   param id - uniquely identifies the action, and should correspond to the action property of the menuaction widget.
     *   param runable - function that performs the custom action.
     *   param isAvailable - (optional) - function called when the menu action is about to appear.  Return false to disable the action, or true to enable it.
     *   param isVisible - (optional) - function called when the menu action is about to appear.  Return false to hide the action, or true to make it visible.
     */

    // For example,
    // gc.nav.registerAction('myCustomCloseAction', function() { window.close(); }, function() { return true; }, function() { return true; });

    // Alternatively, to programmatically disable a menu action at any time use:
    // gc.nav.disableAction('myCustomCloseAction);    then enable it again using:  gc.nav.enableAction('myCustomCloseAction');

    gc.nav.registerAction('show_introduction', function () {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 0;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });
    gc.nav.registerAction('show_registers', function () {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 1;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });
    gc.nav.registerAction('show_motor_control', function () {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_vtabcontainer.selectedIndex = 2;
            templateObj.$.ti_widget_menuitem.hideFullMenu();
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });

    gc.nav.registerAction('set_csv_format', function () {
        regFileFormat = "CSV";
        gc.nav.setActionChecked('set_csv_format', true);
        gc.nav.setActionChecked('set_json_format', false);
    }, function () {
        return true; // isAvailable
    }, function () {
        return true; // isVisible
    });
    gc.nav.setActionChecked('set_csv_format', true);

    gc.nav.registerAction('set_json_format', function () {
        regFileFormat = "JSON";
        gc.nav.setActionChecked('set_csv_format', false);
        gc.nav.setActionChecked('set_json_format', true);
    }, function () {
        return true; // isAvailable
    }, function () {
        return true;  // isVisible
    });

    var saveText = function (fname, text, mime) {
        var blob = new Blob([text], {
            type: mime
        });
        saveAs(blob, fname);
        return false;
    };

    var getCsvForRegs = function () {
        var result = "register,address,value\n";
        var regPage = document.querySelector('#ti_widget_register_page');
        var regBlocks = regPage && regPage.regBlocks;
        if (regBlocks) {
            for (var b = 0; b < regBlocks.length; b++) {
                var regBlock = regBlocks[b];
                for (var r = 0; r < regBlock.registers.length; r++) {
                    var reg = regBlock.registers[r];
                    result += reg.name + "," + reg.addr + "," + reg.value + "\n";
                }
            }
        }
        return result;
    };

    /*
     * saveRegs saves the register values from the register page into either a CSV file 
     * format or a stringified JSON object and saves them to a file.  If isSaveAs
     * parameter is true (or the saveAsFilePath global has not been set), 
     * it shows the 'browseAndSave' dialog to allow the user to 
     * specify which folder to save the file into, otherwise it saves the contents to 
     * the file path specified by saveAsFilePath
     */
    var saveRegs = function (calledBy, isSaveAs) {
        var fileFilter;
        var regData;
        switch (regFileFormat) {
            case "CSV":
                regData = getCsvForRegs();
                fileFilter = ".csv";
                // make sure the file extension matches the regFileFormat
                if (saveAsFilePath) {
                    saveAsFilePath = saveAsFilePath.replace(".json", ".csv");
                }
                break;
            case "JSON":
            default:
                fileFilter = ".json";
                regData = JSON.stringify(templateObj.getRegisterData());
                // make sure the file extension matches the regFileFormat
                if (saveAsFilePath) {
                    saveAsFilePath = saveAsFilePath.replace(".csv", ".json");
                }
        }
        var fileName = saveAsFilePath;
        if (!fileName) {
            fileName = "regs" + fileFilter;
        }
        if (gc.desktop.isDesktop()) {
            if (isSaveAs || !saveAsFilePath) {
                gc.File.browseAndSave(regData, fileName, fileFilter, function (result, errorInfo) {
                    if (errorInfo) {
                        templateObj.logMsgAndNotify("error", calledBy + " " + errorInfo, false, true);
                    } else if ((result) && (result.localPath)) {
                        saveAsFilePath = result.localPath;
                        gc.nav.enableAction('save_registers');
                        templateObj.$.ti_widget_eventlog_view.log("info", "Registers saved to " + result.localPath);
                    }
                });
            } else {
                gc.File.save(regData, { "localPath": saveAsFilePath }, null, function (errorInfo) {
                    if (errorInfo) {
                        templateObj.logMsgAndNotify("error", "Save Registers As: " + errorInfo, false, true);
                    } else {
                        templateObj.$.ti_widget_eventlog_view.log("info", "Registers saved to " + saveAsFilePath);
                    }
                });
            }
        } else {
            saveText(fileName, regData, 'text/plain;charset=utf-8');
            templateObj.$.ti_widget_eventlog_view.log("info", calledBy + ": " + fileName + " file downloaded");
        }
    };

    /*
     * getRegsFromCsv decodes the string passed in as a parameter as the contents of a CSV file
     * and updates the registers in the regBlocks object with the values from the CSV file.
     * NOTE that the target is not updated with these values by this function.  To update the
     * target registers, please call setRegisterData.
     */
    getRegsFromCsv = function (csvContents) {
        var lines = csvContents.split('\n');
        lines.shift(); // discard the title row
        // create a map of the csv values
        var csvMap = new Map();
        for (var i = 0; i < lines.length; i++) {
            var regInfo = lines[i].split(",");
            var name = regInfo[0].trim().toLowerCase();
            var addr = +regInfo[1];
            var value = +regInfo[2];
            var key = name + "@" + addr;
            csvMap.set(key, value);
        }
        var regPage = document.querySelector('#ti_widget_register_page');
        var regBlocks = regPage && regPage.regBlocks;
        if (regBlocks) {
            for (var b = 0; b < regBlocks.length; b++) {
                var regBlock = regBlocks[b];
                for (var r = 0; r < regBlock.registers.length; r++) {
                    var reg = regBlock.registers[r];
                    var regAdrs = +reg.addr;
                    var regKey = reg.name.trim().toLowerCase() + "@" + regAdrs;
                    if (csvMap.has(regKey)) {
                        reg.value = csvMap.get(regKey);
                    }
                }
            }
        }
        return regBlocks;
    };

    gc.nav.registerAction('load_registers', function () {
        if ((templateObj) && (templateObj.$)) {
            gc.File.browseAndLoad(null, {
                bin: false
            }, function (contents, properties) {
                var regData = null;
                try {
                    var fileType = regFileFormat;
                    var dotIndex = (properties && properties.name) ? properties.name.lastIndexOf(".") : -1;
                    if (dotIndex >= 0 && dotIndex < properties.name.length) {
                        fileType = properties.name.substring(dotIndex + 1).toUpperCase();
                        if (fileType !== "JSON" && fileType !== "CSV") {
                            fileType = regFileFormat;
                        }
                    }
                    switch (fileType) {
                        case "CSV":
                            regData = getRegsFromCsv(contents);
                            break;
                        default:
                        case "JSON":
                            regData = JSON.parse(contents);
                            if (typeof regData === "string") {
                                regData = JSON.parse(regData);
                            }
                            if (regData && regData.data) {
                                regData = regData.data;
                            }
                            break;
                    }
                    if (regData) {
                        if (templateObj.setRegisterData(regData)) {
                            templateObj.$.ti_widget_eventlog_view.log("info", "Load Registers: loaded from :" + properties.localPath);
                        }
                    }
                } catch (ex) {
                    templateObj.logMsgAndNotify("error", "Load Registers: Could not parse " + regFileFormat + " file:" + ex, false, true);
                }

            });
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });

    gc.nav.registerAction('save_registers', function () {
        if ((templateObj) && (templateObj.$)) {
            saveRegs("Save Registers", false);
        }
    }, function () {
        return saveAsFilePath;  // isAvailable
    }, function () {
        return saveAsFilePath;  // isVisible
    });
    gc.nav.disableAction('save_registers');
    gc.nav.registerAction('save_registers_as', function () {
        if ((templateObj) && (templateObj.$)) {
            saveRegs("Save Registers As", true);
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });

    gc.nav.registerAction('open_log_pane', function () {
        if ((templateObj) && (templateObj.$)) {
            templateObj.$.ti_widget_eventlog_view.openView();
        }
    }, function () {
        return true;
    }, function () {
        return true;
    });

    gc.nav.registerAction('open_scripting_window', function () {
        window.open('app/scripting.html', '_evm_scripting');
    }, function () {
        return true;
    }, function () {
        return true;
    });
});

/*
 *  Boilerplate code for working with components in the application gist
 */


var initComplete = false;
var templateObj;
var registerModel;


// Wait for DOMContentLoaded event before trying to access the application template
var init = function () {
    templateObj = document.querySelector('#template_obj');

    // Wait for the template to fire a dom-change event to indicate that it has been 'stamped'
    // before trying to access components in the application.
    templateObj.addEventListener('dom-change', function () {
        if (initComplete) return;
        this.async(function () {
            initComplete = true;
            console.log("Application template has been stamped.");
            templateObj.$.ti_widget_toast.hideToast();
            templateObj.$.ti_widget_eventlog_view.log("info", "Application started.");
            gc.nav.registerChecklistActions('fileFormat.', fileFormat,
                {
                    onSelectionHandler: function (formatName, detail) {
                        regFileFormat = formatName;
                        gc.localStorage.setItem('fileFormat', formatName);
                    },

                    onIsAvailableHandler: function () {
                        return false;
                    }
                }
                , gc.localStorage.getItem('fileFormat') || regFileFormat, false);
            registerModel = Polymer.dom(document).querySelector('ti-model-register');
            if (registerModel) {
                registerModel.addEventListener("usbHidModelReceivingData", function (event) {
                    templateObj.logMsgAndNotify("debug", "Receiving data from target.", false, false);
                });
            } else {
                console.log("no ti-model-register component found in this application!")
            }

            // Now that the template has been stamped, you can use 'automatic node finding' $ syntax to access widgets.
            // e.g. to access a widget with an id of 'widget_id' you can use templateObj.$.widgetId
            gc.databind.registry.bind('ti_widget_droplist_auto_read.selectedValue', 'ti_widget_register_page.autoReadValue');
            gc.databind.registry.getBinding('reg.STATUS').addStreamingListener({
                onDataReceived: function (value) {
                    let lo = gc.databind.registry.getBinding('reg.TLB').getValue();
                    let hi = gc.databind.registry.getBinding('reg.THB').getValue();
                    let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                    let TMP = ((hi << 8) | lo); // * 2 ** -7;
                    TMP = (TMP > 0x7FFF) ? TMP - 0x10000 : TMP;
                    TMP *= range ? (2 ** -7) : (2 ** -4);
                    myPlotly.update(TMP);
                    templateObj.$.ti_widget_markdown_numeral.markdownText = '<h1>' + TMP.toFixed(2) + '°C</h1>';
                    var d = new Date();
                    timestamp = d.getFullYear() + '/' + (d.getMonth() + 1) + '/' + d.getDate() + ' ' + d.getHours() + ':' + d.getMinutes() + ':' + d.getSeconds();
                    FileSaver.saveData(TMP + ',' + timestamp + '\n');
                }
            });

            const glob = {
                offset : {
                    updateHex() {
                        let lo = gc.databind.registry.getBinding('reg.TEMP_OFFSET_LSB').getValue();
                        let hi = gc.databind.registry.getBinding('reg.TEMP_OFFSET_MSB').getValue();
                        let x = (hi << 8) + lo;
                        let y = '0x' + x.toString(16).toUpperCase();
                        templateObj.$.ti_widget_input_offset_hex.value = y;
                        console.log("updateHex", y);
                        let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                        x = (x > 0x7FFF) ? x - 0x10000 : x;
                        x *= range ? (2 ** -7) : (2 ** -4);
                        templateObj.$.ti_widget_input_offset.value = x;
                    }
                },
                alertLo : {
                    update() {
                        let lo = gc.databind.registry.getBinding('reg.ALERT_LOW_LSB').getValue();
                        let hi = gc.databind.registry.getBinding('reg.ALERT_LOW_MSB').getValue();
                        let x = (hi << 8) + lo;
                        let y = '0x' + x.toString(16).toUpperCase();
                        templateObj.$.ti_widget_input_alert_lo_hex.value = y;
                        console.log("update", y);
                        let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                        x = (x > 0x7FFF) ? x - 0x10000 : x;
                        x *= range ? (2 ** -7) : (2 ** -4);
                        templateObj.$.ti_widget_input_alert_lo.value = x;
                    }
                },
                alertHi : {
                    update() {
                        let lo = gc.databind.registry.getBinding('reg.ALERT_HIGH_LSB').getValue();
                        let hi = gc.databind.registry.getBinding('reg.ALERT_HIGH_MSB').getValue();
                        let x = (hi << 8) + lo;
                        let y = '0x' + x.toString(16).toUpperCase();
                        templateObj.$.ti_widget_input_alert_hi_hex.value = y;
                        console.log("update", y);
                        let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                        x = (x > 0x7FFF) ? x - 0x10000 : x;
                        x *= range ? (2 ** -7) : (2 ** -4);
                        templateObj.$.ti_widget_input_alert_hi.value = x;
                    }
                }
            }
            gc.databind.registry.getBinding('reg.ALERT_LOW_LSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("LSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.alertLo.update();
                }
            });
            gc.databind.registry.getBinding('reg.ALERT_LOW_MSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("MSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.alertLo.update();
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_alert_lo_hex.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("hex", oldValue.toString(16), newValue.toString(16));
                    let z = parseInt(newValue, 16);
                    let hi = z >> 8;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.ALERT_LOW_MSB');
                    let bl = gc.databind.registry.getBinding('reg.ALERT_LOW_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_alert_lo.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("alertLo", oldValue, newValue);
                    let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                    let x = Number(newValue);
                    let z = x < 0 ? x + 0x10000 : x;
                    z *= range ? (2 ** 7) : (2 ** 4);
                    let hi = (z >> 8) & 0xFF;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.ALERT_LOW_MSB');
                    let bl = gc.databind.registry.getBinding('reg.ALERT_LOW_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });

            gc.databind.registry.getBinding('reg.ALERT_HIGH_LSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("LSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.alertHi.update();
                }
            });
            gc.databind.registry.getBinding('reg.ALERT_HIGH_MSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("MSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.alertHi.update();
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_alert_hi_hex.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("hex", oldValue.toString(16), newValue.toString(16));
                    let z = parseInt(newValue, 16);
                    let hi = z >> 8;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.ALERT_HIGH_MSB');
                    let bl = gc.databind.registry.getBinding('reg.ALERT_HIGH_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_alert_hi.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("alertHi", oldValue, newValue);
                    let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                    let x = Number(newValue);
                    let z = x < 0 ? x + 0x10000 : x;
                    z *= range ? (2 ** 7) : (2 ** 4);
                    let hi = (z >> 8) & 0xFF;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.ALERT_HIGH_MSB');
                    let bl = gc.databind.registry.getBinding('reg.ALERT_HIGH_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });

            gc.databind.registry.getBinding('reg.TEMP_OFFSET_LSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("LSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.offset.updateHex();
                }
            });
            gc.databind.registry.getBinding('reg.TEMP_OFFSET_MSB').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("MSB", (oldValue || '').toString(16), newValue.toString(16));
                    glob.offset.updateHex();
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_offset_hex.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("hex", oldValue.toString(16), newValue.toString(16));
                    let z = parseInt(newValue, 16);
                    let hi = z >> 8;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.TEMP_OFFSET_MSB');
                    let bl = gc.databind.registry.getBinding('reg.TEMP_OFFSET_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });
            gc.databind.registry.getBinding('widget.ti_widget_input_offset.value').addChangedListener({
                onValueChanged: function (oldValue, newValue) {
                    console.log("offset", oldValue, newValue);
                    let range = gc.databind.registry.getBinding('reg.CFG1.TEMP_FMT').getValue();
                    let x = Number(newValue);
                    let z = x < 0 ? x + 0x10000 : x;
                    z *= range ? (2 ** 7) : (2 ** 4);
                    let hi = (z >> 8) & 0xFF;
                    let lo = z & 0xFF;
                    let bh = gc.databind.registry.getBinding('reg.TEMP_OFFSET_MSB');
                    let bl = gc.databind.registry.getBinding('reg.TEMP_OFFSET_LSB');
                    bh.setValue(hi);
                    bl.setValue(lo);
                }
            });

            const sendConvertTemp = function () {
                const u2a = document.querySelector('ti-model-register').getCodec();
                u2a.uartInterface.uartCodec.OWTransfer([0xCC, 0x44], 0);
            };

            const updatePlot = () => {
                const clientRect = templateObj.$.ti_widget_plot.getBoundingClientRect();
                templateObj.$.ti_widget_plot.layout.width = clientRect.width;
                templateObj.$.ti_widget_plot.layout.height = clientRect.height;
                templateObj.$.ti_widget_plot.redraw();
            };
            new ResizeObserver(updatePlot).observe(templateObj.$.ti_widget_plot);

            const setChartRange = function () {
                const range = [templateObj.$.ti_widget_input_chart_min.value, templateObj.$.ti_widget_input_chart_max.value]
                templateObj.$.ti_widget_plot.layout.yaxis.autorange = false;
                templateObj.$.ti_widget_plot.layout.yaxis.range = range;
            };
            templateObj.$.ti_widget_input_chart_min.addEventListener("change", setChartRange);
            templateObj.$.ti_widget_input_chart_max.addEventListener("change", setChartRange);

            if (!(isOnline | dbgOnline))
                // https://nwjs.readthedocs.io/en/nw13/References/File%20Dialogs/#attribute-nwsaveas
                templateObj.$.file_input.addEventListener('change', e => {
                    FileSaver.saveFile = templateObj.$.file_input.value;
                    templateObj.$.file_input.value = "";
                    FileSaver.saveLog = true;
                });
            templateObj.$.ti_widget_button_saveStart.addEventListener("click", e => FileSaver.saveStart());
            templateObj.$.ti_widget_button_saveStop.addEventListener("click", e => FileSaver.saveStop());
            templateObj.$.ti_widget_toggle_sendConvertTemp.addEventListener("change", e => {
                const u2a = document.querySelector('ti-model-register').getCodec();
                u2a.uartInterface.uartCodec.sendConvertTemp = templateObj.$.ti_widget_toggle_sendConvertTemp.checked;
            });

            // const scratchAction = (cmd, start, stop, bytes, read) => {
            //     const u2a = document.querySelector('ti-model-register').getCodec();
            //     templateObj.$.ti_widget_markdown_scratch2.markdownText = "|Addr1|Addr2|1|2|3|4|5|6|7|8|\n";
            //     templateObj.$.ti_widget_markdown_scratch2.markdownText += "|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|\n";

            //     for (let addr = start; addr <= stop; addr++) {
            //         if (u2a)
            //             u2a.uartInterface.uartCodec.scratchpad2([0xCC, cmd, (addr >> 8), (addr & 0xFF)], read).then(b => {
            //                 console.log("resolve", b);
            //                 templateObj.$.ti_widget_markdown_scratch2.markdownText += '|' + b.map(x => '0x' + (x).toString(16).padStart(2, '0').toUpperCase()).join('|') + '|\n';
            //             });
            //     }
            // }
            const scratch2 = {
                preview : "",
                init() {
                    templateObj.$.ti_widget_markdown_scratch2.markdownText = "|Addr1|Addr2|1|2|3|4|5|6|7|8|\n";
                    templateObj.$.ti_widget_markdown_scratch2.markdownText += "|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|\n";
                    scratch2.preview = "Addr1,Addr2,1,2,3,4,5,6,7,8\n";
                },
                append(b) {
                    templateObj.$.ti_widget_markdown_scratch2.markdownText += '|' + b.map(x => '0x' + (x).toString(16).padStart(2, '0').toUpperCase()).join('|') + '|\n';
                    scratch2.preview += b.map(x => '0x' + (x).toString(16).padStart(2, '0').toUpperCase()).join(',') + '\n';
                },
                readFile(write = false, commit = false) {
                    const u2a = document.querySelector('ti-model-register').getCodec();
                    if ((templateObj) && (templateObj.$)) {
                        gc.File.browseAndLoad(null, {
                            bin: false
                        }, function (contents, properties) {
                            try {
                                scratch2.init();
                                let lines = contents.split("\n");
                                lines.shift(); //discard label row
                                for (var i = 0; i < lines.length; i++) {
                                    let cells = lines[i].split(",").map(x => parseInt(x));
                                    if (cells.length == 10) {
                                        scratch2.append(cells);
                                        if (write)
                                            u2a.uartInterface.uartCodec.OWTransfer([0xCC, 0x0F, ...cells], 1);
                                        if (write && commit)
                                            u2a.uartInterface.uartCodec.OWTransfer([0xCC, 0x55, 0xA5], 0);
                                    }
                                    else
                                        templateObj.logMsgAndNotify("error", "Read file ignored line " + i + ": \"" + lines[i] + "\"", true, false);
                                }
                            } catch (ex) {
                                templateObj.logMsgAndNotify("error", "Could not parse file:" + ex, false, true);
                            }
                        });
                    }
                }
            }
            templateObj.$.ti_widget_button_readScratch2.addEventListener("click", e => {
                const u2a = document.querySelector('ti-model-register').getCodec();
                let start = parseInt(templateObj.$.ti_widget_input_scratch2_read_start.value);
                let stop = parseInt(templateObj.$.ti_widget_input_scratch2_read_stop.value);
                start &= 0xFFF8;
                stop &= 0xFFF8;
                scratch2.init();
                for (let addr = start; addr <= stop; addr += 8) {
                    if (u2a)
                        u2a.uartInterface.uartCodec.scratchpad2([0xCC, 0xAA, (addr >> 8), (addr & 0xFF)], 9).then(b => {
                            console.log("resolve", b);
                            scratch2.append(b);
                        });
                }
            });
            templateObj.$.ti_widget_button_readEEPROM.addEventListener("click", e => {
                const u2a = document.querySelector('ti-model-register').getCodec();
                scratch2.init();
                let start = parseInt(templateObj.$.ti_widget_input_scratch2_read_start.value);
                let stop = parseInt(templateObj.$.ti_widget_input_scratch2_read_stop.value);
                start &= 0xFFF8;
                stop &= 0xFFF8;
                for (let addr = start; addr <= stop; addr += 8) {
                    if (u2a) {
                        u2a.uartInterface.uartCodec.scratchpad2([0xCC, 0xF0, (addr >> 8), (addr & 0xFF)], 0)
                        u2a.uartInterface.uartCodec.scratchpad2([], 8, false).then(b => {
                            console.log("resolve", b);
                            scratch2.append([addr >> 8, addr & 0xFF].concat(b));
                        });
                    }
                }
            });
            const scratch2WriteBlock = (commit = false) => {
                const u2a = document.querySelector('ti-model-register').getCodec();
                let addr1 = parseInt(templateObj.$.ti_widget_input_scratch2_write_addr1.value);
                let addr2 = parseInt(templateObj.$.ti_widget_input_scratch2_write_addr2.value);
                addr2 &= 0xF8;
                let d = [0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08];
                d[0] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte1.value);
                d[1] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte2.value);
                d[2] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte3.value);
                d[3] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte4.value);
                d[4] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte5.value);
                d[5] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte6.value);
                d[6] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte7.value);
                d[7] = parseInt(templateObj.$.ti_widget_input_scratch2_write_byte8.value);
                u2a.uartInterface.uartCodec.OWTransfer([0xCC, 0x0F, addr1, addr2, ...d], 1);
                if (commit)
                    u2a.uartInterface.uartCodec.OWTransfer([0xCC, 0x55, 0xA5], 0);
            }
            templateObj.$.ti_widget_button_writeScratch2.addEventListener("click", e => {
                scratch2WriteBlock();
            });
            templateObj.$.ti_widget_button_commitScratch2.addEventListener("click", e => {
                scratch2WriteBlock(true);
            });
            templateObj.$.ti_widget_button_scratch2_readFile.addEventListener("click", e => {
                scratch2.readFile();
            });
            templateObj.$.ti_widget_button_scratch2_commitFromFile.addEventListener("click", e => {
                scratch2.readFile(true,true);
            });
            templateObj.$.ti_widget_button_scratch2_saveFile.addEventListener("click", e => {
                let data = scratch2.preview;
                let fileFilter = ".csv";
                let fileName = "eeprom" + fileFilter;
                if (gc.desktop.isDesktop()) {
                    gc.File.browseAndSave(data, fileName, fileFilter, function (result, errorInfo) {
                        if (errorInfo) {
                        } else if ((result) && (result.localPath)) {
                        }
                    });
                } else {
                    var blob = new Blob([data], {
                        type: 'text/plain;charset=utf-8'
                    });
                    saveAs(blob, fileName);
                }
            });

            // open the eventlog viewer when the user clicks on the book icon in the status bar:
            templateObj.$.ti_widget_eventlog_view.closeView(); // ensure that view states are in sync with application
            templateObj.$.ti_widget_statusbar.addEventListener("status-icon-clicked", function (event) {
                // TODO: add whatever actions you wish to perform in response to the appStatusIcon in the statusbar
                // being clicked.  See event.details for info on the icon and app status
            });

            // Expand vtabcontainer nav bar when user clicks on menu icon or 'Menu' label
            templateObj.toggleMenu = function (event) {
                templateObj.$.ti_widget_vtabcontainer.isExpanded = !templateObj.$.ti_widget_vtabcontainer.isExpanded;
            };
            templateObj.$.ti_widget_icon_button_menu.addEventListener('click', templateObj.toggleMenu);
            templateObj.$.ti_widget_label_menu.addEventListener('click', templateObj.toggleMenu);
            var regPage = document.querySelector('#ti_widget_register_page');
            if (regPage) {
                // get the active device name from the register page
                templateObj.$.ti_widget_menubar.productName = regPage.deviceName;

                // called when the user clicks the Write Register button:
                regPage.addEventListener("write_reg", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));

                // called when the user clicks the Write All Registers button:
                regPage.addEventListener("write_regs", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));

                // called when the user clicks the Read Register button:
                regPage.addEventListener("read_reg", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));

                // called when the user clicks the Read All Registers button:
                regPage.addEventListener("read_regs", function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify(detailsObj.statusType, detailsObj.statusMsg, false, true);
                }.bind(this));

                // called when the user changes the immediate / deferred mode
                regPage.addEventListener("update_mode_changed", function (event) {
                    var detailsObj = event.detail;
                    templateObj.$.ti_widget_eventlog_view.log("info", "request to change update mode to " + detailsObj.mode.label);
                }.bind(this));

                // called when the user changes the auto read refresh rate
                regPage.addEventListener("auto_read_config_changed", function (event) {
                    var detailsObj = event.detail;
                    templateObj.$.ti_widget_eventlog_view.log("info", "request to change automatic read interval to " + detailsObj.autoRead.label);
                }.bind(this));
            }

            var usbHidService = gc.services['ti-service-usbhid'];
            if (usbHidService) {
                usbHidService.addEventListener('usbHidError', function (event) {
                    var detailsObj = event.detail;
                    templateObj.logMsgAndNotify("error", detailsObj.info + ": " + detailsObj.error, false, true);
                })
            }

            templateObj.getRegisterData = function () {
                var regData = {
                    signature: "register-data",
                    data: []
                };
                var regPage = document.querySelector('#ti_widget_register_page');
                var regBlocks = regPage && regPage.regBlocks;
                if (regBlocks) {
                    for (var i = 0; i < regBlocks.length; i++) {
                        var registersData = [];
                        for (var r = 0; r < regBlocks[i].registers.length; r++) {
                            var reg = regBlocks[i].registers[r];
                            var regName = reg.name.toLowerCase();
                            regName = regName.replace(/&/g, "_");
                            var item = {
                                idx: r,
                                id: regName,
                                value: reg.value
                            };
                            registersData.push(item);
                        }
                        regData.data.push(registersData);
                    }
                }
                return regData;
            };

            templateObj.setRegisterData = function (jsonObj) {
                var ok = true;
                var regData = jsonObj;
                var regPage = document.querySelector('#ti_widget_register_page');
                var regBlocks = regPage && regPage._registerView.regBlocks;
                if (regBlocks) {
                    for (var b = 0; b < Math.min(regData.length, regBlocks.length); b++) {
                        var idx = 0;
                        var dRegName = "";
                        var dBlock = regData[b];
                        var dNumRegs = dBlock.length;
                        if (dBlock.registers) {
                            dNumRegs = dBlock.registers.length;
                        }
                        for (var r = 0; r < regBlocks[b].registers.length; r++) {
                            var reg = regBlocks[b].registers[r];
                            var regName = reg.name.toLowerCase();
                            var dReg = null;
                            if (dBlock.registers && dBlock.registers.length > idx) {
                                // from .csv file
                                dReg = dBlock.registers[idx];
                                dRegName = dReg.name;
                            } else if (dBlock.length && dBlock.length > idx) {
                                // from .json file
                                dReg = dBlock[idx];
                                dRegName = dReg.id;
                            }
                            if (dRegName && dRegName.toLowerCase() === regName) {
                                idx++;
                                reg.value = dReg.value;
                                if (!regPage.isDeferredWrite) {
                                    reg.valueLastWritten = reg.value;
                                    var intValue = gc.utils.getValueFromHexString("" + reg.value);
                                    regPage.writeValueToTargetRegister(intValue, null, b, r);
                                } else {
                                    regPage._registerView.refreshRegValue(b, r);
                                }
                            }
                        }
                        if (idx < dNumRegs) {
                            var msg = "load registers failed: register name mismatch (" + dRegName + " not found.)";
                            templateObj.logMsgAndNotify("warning", msg, false, true);
                            ok = false;
                            break;
                        }
                    }
                    if (regPage.isDeferredWrite) {
                        regPage._registerView._deferredAccessCtr++; // update the page
                    }
                    regPage.refreshGrid();
                }
                return ok;
            };

            templateObj.showToastMsg = function (msgType, msg, durationInSeconds) {
                templateObj.$.ti_widget_toast.hideToast();
                templateObj.$.ti_widget_toast.message = msg;
                switch (msgType.toLowerCase()) {
                    case "error":
                        templateObj.$.ti_widget_toast.iconName = "error";
                        templateObj.$.ti_widget_toast.backgroundColor = "red";
                        break;
                    case "warning":
                        templateObj.$.ti_widget_toast.iconName = "warning";
                        templateObj.$.ti_widget_toast.backgroundColor = "orange";
                        break;
                    default:
                        templateObj.$.ti_widget_toast.iconName = "check";
                        templateObj.$.ti_widget_toast.backgroundColor = "teal";
                        break;
                }
                templateObj.$.ti_widget_toast.duration = +durationInSeconds;
                templateObj.$.ti_widget_toast.showToast();
                console.log("toast msg [" + msgType + "]: " + msg);
            };

            templateObj.logMsgAndNotify = function (msgType, msg, showInStatusBar, showInToast) {
                var iconName = templateObj.$.ti_widget_eventlog_view.iconForItem(msgType);
                if (showInStatusBar) {
                    var statusBar = document.querySelector('ti-widget-statusbar');
                    if (statusBar) {
                        statusBar.appStatusText = msg;
                        statusBar.appStatusTooltip = msg;
                        statusBar.setIcon(iconName);
                    }
                }
                if (showInToast) {
                    templateObj.showToastMsg(msgType, msg, 4);
                }
                templateObj.$.ti_widget_eventlog_view.log(msgType, msg);
            };
            if (templateObj.$.ti_widget_label_learnmore) {
                templateObj.$.ti_widget_label_learnmore.addEventListener("click", function (event) {
                    templateObj.$.ti_widget_label_productname.fontColor = "#444";
                    templateObj.$.ti_widget_label_productdesc.fontColor = "#444";
                    templateObj.$.ti_widget_label_learnmore.setAttribute('hidden', true);
                    templateObj.$.ti_widget_label_showless.removeAttribute('hidden');
                    templateObj.$.ti_widget_vtab_home.style.backgroundColor = "white";
                    templateObj.$.ti_widget_tabcontainer.selectedIndex = 1;
                });
            }
            if (templateObj.$.ti_widget_label_showless) {
                templateObj.$.ti_widget_label_showless.addEventListener("click", function (event) {
                    templateObj.$.ti_widget_label_productname.fontColor = "white";
                    templateObj.$.ti_widget_label_productdesc.fontColor = "white";
                    templateObj.$.ti_widget_label_learnmore.removeAttribute('hidden');
                    templateObj.$.ti_widget_label_showless.setAttribute('hidden', true);
                    templateObj.$.ti_widget_vtab_home.style.backgroundColor = "rgb(20, 140, 156)";
                    templateObj.$.ti_widget_tabcontainer.selectedIndex = 0;
                });
            }
            if (templateObj.$.ti_widget_button_connect) {
                templateObj.$.ti_widget_button_connect.addEventListener("click", function (event) {
                    if (gc.connectionManager) {
                        gc.connectionManager.connect();
                    }
                });
            }


        }, 1);

    });
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
    init();
} else {
    document.addEventListener('DOMContentLoaded', init);
}